from IPython.display import display, clear_output
import ipywidgets as widgets
from clease_gui.base_dashboard import BaseDashboard
from .emt_calculator import EMTDashboard

__all__ = ["CalculateDashboard"]


class CalculateDashboard(BaseDashboard):
    def initialize(self):
        self.emt_out = widgets.Output()
        emt_dash = EMTDashboard(self.app_data)
        with self.emt_out:
            clear_output()
            emt_dash.display()

    def display(self):
        display(self.emt_out)



"""Adaptation of the ASE NGL viewer"""
import logging
from IPython.display import display
import nglview as nv
import ase

from clease_gui import register_logger

logger = logging.getLogger(__name__)
register_logger(logger)


class MCNGLDisplay:
    def __init__(self, xsize=500, ysize=500):
        self.xsize = xsize
        self.ysize = ysize
        self.make_viewer()

    def display(self):
        display(self.viewer)

    def make_viewer(self, atoms: ase.Atoms = None):
        if atoms is not None:
            s = nv.ASEStructure(atoms)
            self.viewer = nv.widget.NGLWidget(s)
        else:
            self.viewer = nv.widget.NGLWidget()
            return
        # Remove the bond cylinders between atoms
        self.viewer.clear_representations()

        self.viewer._set_size(self.xsize, self.ysize)
        self.viewer.camera = "orthographic"

        # self.viewer.add_unitcell()
        self.viewer.add_spacefill()
        self.viewer.update_spacefill(
            radiusType="covalent",
            radiusScale=0.5,
            color_scheme="element",
            color_scale="rainbow",
        )

    def get_orientation(self):
        return self.viewer._camera_orientation

    def replace_structure(self, atoms: ase.Atoms) -> None:
        ngl_traj = nv.ASEStructure(atoms)
        struct = dict(data=ngl_traj.get_structure_string(), ext="pdb")
        self.viewer._remote_call(
            "replaceStructure",
            target="Widget",
            args=[
                struct,
            ],
        )

    def update_coordinates(self, atoms: ase.Atoms) -> None:
        self.viewer.set_coordinates({0: atoms.positions})



"""Open a Jupyter notebook"""

import click
import shutil
from pathlib import Path, PurePath
from clease_gui.utils import get_assets_path
from .main_cli import clease_gui_cli

__all__ = []

DEFAULT_FILENAME = "clease_gui.ipynb"


def copy_default_notebook(filename: PurePath) -> None:
    src = get_assets_path() / "default_notebook.ipynb"
    if filename.exists():
        raise RuntimeError(f"File {filename} already exists.")
    shutil.copy(src, filename)


@clease_gui_cli.command()
@click.option(
    "-f",
    "--filename",
    type=str,
    default=DEFAULT_FILENAME,
    required=False,
    help=f"""
Create a new notebook to be used with the CLEASE GUI.
Defaults to "{DEFAULT_FILENAME}".""",
)
def new(filename: str) -> None:
    """Create a mew notebook to be used with the CLEASE GUI."""
    file_path = Path(filename)
    # Ensure we're making an .ipynb file
    file_path = file_path.with_suffix(".ipynb")
    if file_path.exists():
        click.echo(f'File "{file_path}" already exists.')
        return
    click.echo(f"Creating new notebook: {file_path}")
    copy_default_notebook(file_path)



from collections import namedtuple
from contextlib import contextmanager
import functools
import ipywidgets as widgets
from IPython.display import display

from clease_gui.base_dashboard import BaseDashboard
import clease_gui.colors as colors

__all__ = ["StatusBar", "update_statusbar"]

UpdatableWidget = namedtuple("UpdatableWidget", "widget, update_function")


def update_statusbar(func):
    """Decorator to update statusbar after function completion, and mark
    status as busy from a dashboard object."""

    @functools.wraps(func)
    def _wrapper(self, *args, **kwargs):
        try:
            statusbar: StatusBar = self.app_data[self.KEYS.STATUS]["statusbar"]
        except KeyError:
            # Something happened, we couldn't get the statusbar
            return func(self, *args, **kwargs)

        # Set statusbar as busy
        with statusbar.status_busy():
            return_val = func(self, *args, **kwargs)
        return return_val

    return _wrapper


class StatusBar(BaseDashboard):
    IS_BUSY = "is_busy"

    def __init__(self, *args, **kwargs):
        self._processes = 0
        self.all_widgets = {}
        super().__init__(*args, **kwargs)
        # Register this statusbar in the app data
        # So others can register it as busy
        self.app_data[self.KEYS.STATUS] = {"statusbar": self}
        self._subscribe()

    def _subscribe(self) -> None:
        """Subscribe the status widgets to the app data"""

        def _wrapper(fnc):
            def _dec(change):
                return fnc()

            return _dec

        # Wrap the updater, since it doesn't care about the notification.
        wrapped = _wrapper(self.update_all)
        for key in [self.KEYS.SETTINGS, self.KEYS.ECI, self.KEYS.SUPERCELL]:
            self.app_data.subscribe(key, wrapped)

    def initialize(self):
        # Create widgets
        self.status_widget = widgets.HTML(value=self.is_busy_str())
        self.register_status_widget(self.IS_BUSY, self.status_widget, self.is_busy_str)

        self.has_settings_wdgt = widgets.Label(value=self.has_settings())
        self.register_status_widget("settings", self.has_settings_wdgt, self.has_settings)

        self.has_eci_wdgt = widgets.Label(value=self.has_eci())
        self.register_status_widget("eci", self.has_eci_wdgt, self.has_eci)

        self.has_supercell_widget = widgets.Label(value=self.has_supercell())
        self.register_status_widget("supercell", self.has_supercell_widget, self.has_supercell)

    def display(self):
        hbox1 = widgets.VBox(
            children=[
                self.status_widget,
                self.has_settings_wdgt,
            ],
            # Add a bit of margin to the right
            layout=widgets.Layout(margin="0 50px 0 0"),
        )

        hbox2 = widgets.VBox(
            children=[
                self.has_eci_wdgt,
                self.has_supercell_widget,
            ]
        )

        boxes = widgets.HBox(children=[hbox1, hbox2], layout=widgets.Layout(border="solid 1px"))

        display(boxes)

    def register_status_widget(self, name, widget, func):
        updatable = UpdatableWidget(widget, func)
        self.all_widgets[name] = updatable

    @property
    def processes(self):
        return self._processes

    def has_supercell(self):
        val = self.key_available(self.KEYS.SUPERCELL)
        return self._format_bool("Supercell available?", val)

    def is_busy(self):
        return self.processes > 0

    def is_busy_str(self):
        s = "Status:   "
        if self.is_busy() > 0:
            # status = Colors.latex_red('Busy')
            status = colors.Colors.html_color("Busy", "red")
        else:
            status = colors.Colors.html_color("Idle", "green")
        return f"{s} {status}"

    def has_settings(self):
        val = self.key_available(self.KEYS.SETTINGS)
        return self._format_bool("Settings available?", val)

    def has_eci(self):
        val = self.key_available(self.KEYS.ECI)
        return self._format_bool("ECI available?", val)

    def key_available(self, key: str) -> bool:
        """A key is considered missing if it is either None or
        doesn't exist in the app data."""
        return self.app_data.get(key, None) is not None

    def _format_bool(self, s, boolean):
        val = colors.bool2symbol[boolean]
        return f"{s}   {val}"

    def update_all(self):
        for updatable_widget in self.all_widgets.values():
            func = updatable_widget.update_function
            updatable_widget.widget.value = func()

    def update_status(self):
        widget = self.all_widgets[self.IS_BUSY]
        func = widget.update_function
        widget.widget.value = func()

    def increment_process(self) -> None:
        self._processes += 1

    def deincrement_process(self) -> None:
        if self._processes < 1:
            raise RuntimeError("Should not be deincrementing process which is less than 1")
        self._processes -= 1

    @contextmanager
    def status_busy(self):
        self.increment_process()
        try:
            self.update_status()
            yield
        finally:
            self.deincrement_process()
            self.update_status()

    def __repr__(self):
        return f"{self.__class__.__name__}(busy={self.is_busy()}, processes={self.processes})"



__all__ = ["Colors", "bool2symbol"]

# True => Green checkmark, False => Big red X
bool2symbol = {True: "\u2705", False: "\N{CROSS MARK}"}


class Colors:
    RED = "\033[31m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    BOLD = "\033[1m"
    ITALIC = "\033[3m"
    RESET = "\033[0m"

    @staticmethod
    def html_color(text, color):
        return f'<font color="{color}">{text}</font>'



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease_gui import register_logger
from clease_gui.base_dashboard import BaseDashboard
from . import CanonicalMC, PlotMCDashboard, MCRunDashboard, MCViewDashboard

__all__ = ["MainMCDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class MainMCDashboard(BaseDashboard):
    def initialize(self):
        self.canonical_mc_out = widgets.Output()
        self.canonical_mc_dashboard = CanonicalMC(self.app_data)
        with self.canonical_mc_out:
            self.canonical_mc_dashboard.display()

        self.plot_out = widgets.Output(layout={"height": "100%"})
        self.plot_dashboard = PlotMCDashboard(self.app_data)
        with self.plot_out:
            self.plot_dashboard.display()

        self.store_mc_run_out = widgets.Output()
        self.store_mc_dashboard = MCRunDashboard(self.app_data)
        with self.store_mc_run_out:
            self.store_mc_dashboard.display()

        self.view_mc_out = widgets.Output()
        self.view_mc_dashboard = MCViewDashboard(self.app_data)
        with self.view_mc_out:
            clear_output()
            self.view_mc_dashboard.display()

        self.tab = widgets.Tab(
            children=[
                self.canonical_mc_out,
                self.plot_out,
                self.store_mc_run_out,
                self.view_mc_out,
            ]
        )

        self.tab.set_title(0, "Canonical MC")
        self.tab.set_title(1, "Plotting")
        self.tab.set_title(2, "MC Runs")
        self.tab.set_title(3, "View MC")

        # A fix for showing the viewer in a tab.
        # Code adapted from aiida.
        # https://github.com/aiidalab/aiidalab-qe/issues/69
        def on_selected_index_change(change):
            index = change["new"]
            # Accessing the viewer only if the corresponding tab is present.
            # i.e. only trigger the toggling if we open the View MC tab.
            if self.tab._titles[str(index)] == "View MC":
                self.view_mc_dashboard.handle_resize()
                viewer = self.view_mc_dashboard.get_ngl_widget()

                def toggle_camera():
                    """Toggle camera between perspective and orthographic."""
                    viewer.camera = (
                        "perspective" if viewer.camera == "orthographic" else "orthographic"
                    )

                toggle_camera()
                toggle_camera()

        self.tab.observe(on_selected_index_change, "selected_index")

    def display(self):
        display(self.tab)



import os
import ipywidgets as widgets
from IPython.display import display
import logging

__all__ = [
    "default_handler",
    "register_logger",
    "OutputWidgetHandler",
    "initialize_clease_gui_logging",
]


class OutputWidgetHandler(logging.Handler):
    """Custom logging handler sending logs to an output widget.
    Class inspired by
    https://ipywidgets.readthedocs.io/en/stable/examples/Output%20Widget.html
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # layout = widgets.Layout(width='100%',
        # height='160px',
        # bor)
        layout = {
            "width": "100%",
            "height": "160px",
            "border": "1px solid black",
            "overflow": "auto",
        }
        self.out = widgets.Output(layout=layout)

    def emit(self, record):
        """Overload of logging.Handler method"""
        formatted_record = self.format(record)
        new_output = {
            "name": "stdout",
            "output_type": "stream",
            "text": formatted_record + os.linesep,
        }
        self.out.outputs = (new_output,) + self.out.outputs

    def show_logs(self):
        """Show the logs"""
        display(self.out)

    def clear_logs(self):
        """Clear the current logs"""
        self.out.clear_output()


default_handler = OutputWidgetHandler()
default_handler.setFormatter(logging.Formatter("%(asctime)s  - [%(levelname)s] %(message)s"))
registered_loggers = []


def register_logger(logger, level=logging.INFO):
    """Register a logger to the default output widget handler.
    Anything emitted by this log will be shown in that widget"""
    global default_handler, registered_loggers
    registered_loggers.append(logger)
    logger.addHandler(default_handler)
    logger.setLevel(level)


def _register_warning_logger():
    """Inspired by
    https://stackoverflow.com/questions/38531786/capturewarnings-set-to-true-doesnt-capture-warnings
    Warning: this changes global variable in logging.
    """
    logging.captureWarnings(True)
    warnings_logger = logging.getLogger("py.warnings")
    register_logger(warnings_logger)


def _register_clease_logger(level=logging.INFO):
    clease_logger = logging.getLogger("clease")
    register_logger(clease_logger, level=level)


def set_all_levels(level) -> None:
    for logger in registered_loggers:
        logger.setLevel(level)


def initialize_clease_gui_logging():
    # _register_clease_logger()
    _register_warning_logger()



# pylint: disable=invalid-name
from pathlib import Path
from packaging.version import parse

__all__ = ("__version__", "version_info")

with Path(__file__).with_name("_version.txt").open("r") as f:
    # A class of type Version
    version_info = parse(f.readline().strip())

__version__ = str(version_info)



import logging
from IPython.display import display
import ipywidgets as widgets

import ase
from ase.optimize import BFGS
from ase.constraints import UnitCellFilter
from ase.calculators.emt import EMT
import clease

from clease_gui import register_logger, utils
from clease_gui.status_bar import update_statusbar
from clease_gui.base_dashboard import BaseDashboard

logger = logging.getLogger(__name__)
register_logger(logger)


class EMTDashboard(BaseDashboard):
    def initialize(self):

        self.calc_button = widgets.Button(description="Calculate with EMT", button_style="primary")
        self.calc_button.on_click(self._on_calc_click)

        self.optimize_cell_widget = widgets.Checkbox(
            description="Optimize cell?",
            value=True,
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.fmax_widget = widgets.BoundedFloatText(
            value=0.02,
            min=1e-4,
            max=999,
            description="Max Force:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        def _on_update_optimize(change):
            if utils.is_value_change(change):
                self._update_widgets()

        self.optimize_cell_widget.observe(_on_update_optimize)
        self._update_widgets()

    def _update_widgets(self):
        # disable fmax if we don't optimize
        fmax_disabled = not self.do_optimize
        self.fmax_widget.disabled = fmax_disabled

    @property
    def fmax(self):
        return self.fmax_widget.value

    @property
    def do_optimize(self):
        return self.optimize_cell_widget.value

    @update_statusbar
    @utils.disable_cls_widget("calc_button")
    def _on_calc_click(self, b):
        with self.event_context(logger=logger) as cm:
            self.calculate()
        if not cm.success:
            logger.warning("Failed to calculate.")

    @property
    def mode(self):
        if self.do_optimize:
            return "Optimization"
        return "Static"

    def calculate(self):
        settings = self.app_data[self.KEYS.SETTINGS]
        if settings is None:
            raise RuntimeError("Need to know the settings to calculate")
        db_name = str(self.get_db_name())
        con = ase.db.connect(settings.db_name)
        num_calculated = 0

        mode = self.mode
        for row in con.select(converged=False):
            logger.info("Calculating row id: %d. Mode: %s", row.id, mode)
            atoms = row.toatoms()
            self._calculate_atoms(atoms)
            custom_kvp = row.key_value_pairs.copy()
            for key in ("struct_type", "name", "converged", "started", "queued"):
                # These keys are set in CLEASE, pop them out
                # pop converged just because we leave that as an "initial" attribute
                custom_kvp.pop(key, None)

            clease.tools.update_db(
                uid_initial=row.id,
                final_struct=atoms,
                db_name=db_name,
                custom_kvp_final=custom_kvp,
            )
            num_calculated += 1
        if num_calculated == 0:
            logger.info("No calculations were performed.")
        else:
            logger.info(f"Completed {num_calculated} calculations.")

    def _calculate_atoms(self, atoms):
        calc = EMT()
        atoms.calc = calc
        if self.do_optimize:
            ucf = UnitCellFilter(atoms)
            opt = BFGS(ucf, logfile=None)
            fmax = self.fmax
            opt.run(fmax=fmax)
        else:
            atoms.get_potential_energy()

    def display(self):
        box = widgets.HBox(children=[self.optimize_cell_widget, self.fmax_widget])
        display(box, self.calc_button)



from typing import List
import logging
import functools
from abc import ABC, abstractmethod
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.style_mixin import WidgetStyleMixin
from clease_gui.app_data import AppDataKeys
from clease_gui import utils, register_logger
from . import structure_gen_kwarg_widgets as kwwidgets

__all__ = ["StructureGenerationDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class StructureGenerationDashboard(BaseDashboard):
    def initialize(self):
        self.scheme_output = widgets.Output()
        self.scheme_widgets = {
            "initial_pool": InitialPoolWidgets(self.app_data, self.scheme_output),
            "random_structure": RandomStructureWidgets(self.app_data, self.scheme_output),
            "ground_state_structure": GroundStateStructure(self.app_data, self.scheme_output),
            "probe_structure": ProbeStructure(self.app_data, self.scheme_output),
        }
        self.available_schemes_b = widgets.Dropdown(
            options=[
                ("Initial Pool", "initial_pool"),
                ("Random Structure", "random_structure"),
                ("Ground-State Structure", "ground_state_structure"),
                ("Probe Structure", "probe_structure"),
            ],
            description="Generation Scheme:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        def on_scheme_change(change):
            if utils.is_value_change(change):
                self.update_scheme_widgets()

        self.available_schemes_b.observe(on_scheme_change)

    def display(self):
        self.update_scheme_widgets()
        display(self.available_schemes_b, self.scheme_output)

    @property
    def current_scheme(self) -> str:
        """Get the current generation scheme"""
        return self.available_schemes_b.value

    def get_scheme_widgets(self) -> "BaseStructGenWidgets":
        scheme = self.current_scheme
        return self.scheme_widgets[scheme]

    def update_scheme_widgets(self):
        widgets = self.get_scheme_widgets()
        widgets.draw_widgets()

    def get_args(self):
        """Get the positional arguments for the scheme"""
        widgets = self.get_scheme_widgets()
        return widgets.make_args()

    def get_kwargs(self):
        """Get the keyword arguments for the scheme"""
        widgets = self.get_scheme_widgets()
        return widgets.make_kwargs()

    def get_generator_name(self):
        widgets = self.get_scheme_widgets()
        return widgets.generator_name()


class BaseStructGenWidgets(WidgetStyleMixin, ABC):
    def __init__(self, app_data, output):
        super().__init__()
        self.app_data = app_data

        self.output = output
        self._widgets_lst = []

        # Mapping from CLEASE kwarg to function which returns its value
        self._clease_kwargs = {}
        # Positional arguments for the CLEASE generator function.
        # Should contain the functions which returns the relevant value
        # in the order they are required.
        self._clease_args = []

    @property
    def widgets_lst(self) -> List[kwwidgets.SimpleWidget]:
        return self._widgets_lst

    @property
    def clease_kwargs(self):
        """Dictionary containing callable functions
        for generating the kwarg inputs.
        """
        return self._clease_kwargs

    @property
    def clease_args(self):
        """List containing callable functions,
        which generates the input arguments"""
        return self._clease_args

    def draw_widgets(self):
        with self.output:
            clear_output(wait=True)
            for widget in self.widgets_lst:
                widget.display()

    @abstractmethod
    def generator_name(self) -> str:
        """Name of the method in the NewStructures class
        to be called for structure generation"""

    def make_args(self):
        """Evaluate the arguments"""
        return tuple(f() for f in self.clease_args)

    def make_kwargs(self) -> dict:
        """Evaluate kwargs"""
        kwargs = {name: f() for name, f in self.clease_kwargs.items()}
        logger.debug("New struct kwargs: %s", kwargs)
        return kwargs

    def register_kwarg_widget(self, *simple_widgets: kwwidgets.SimpleWidget):
        """Function to register a widget for a kwarg.
        One widget may register more than 1 key.
        """

        for simple_widget in simple_widgets:

            for key in simple_widget.widget_dct.keys():
                func = functools.partial(simple_widget.get_value, key)
                self.clease_kwargs[key] = func
            self.widgets_lst.append(simple_widget)

    def register_arg_widget(self, *simple_widgets: kwwidgets.SimpleWidget):
        """Register a widget to return an argument.
        Note that the ordering here matters.
        Can take multiple widgets.
        """
        for simple_widget in simple_widgets:
            for key in simple_widget.widget_dct.keys():
                func = functools.partial(simple_widget.get_value, key)

                self.clease_args.append(func)
            self.widgets_lst.append(simple_widget)


class InitialPoolWidgets(BaseStructGenWidgets):
    def __init__(self, app_data, output):
        widgets = [kwwidgets.AtomsWidget(app_data)]
        super().__init__(app_data, output)
        for widget in widgets:
            self.register_kwarg_widget(widget)

    def generator_name(self) -> str:
        return "generate_initial_pool"


class RandomStructureWidgets(BaseStructGenWidgets):
    def __init__(self, app_data, output):
        widgets = [kwwidgets.AtomsWidget(app_data)]
        super().__init__(app_data, output)
        for widget in widgets:
            self.register_kwarg_widget(widget)

    def generator_name(self) -> str:
        return "generate_random_structures"


class GroundStateStructure(BaseStructGenWidgets):
    def __init__(self, app_data, output):
        key_value_widgets = [
            # kwwidgets.AtomsWidget(app_data),
            kwwidgets.InitTemp(app_data),
            kwwidgets.FinalTemp(app_data),
            kwwidgets.NumTemp(app_data),
            kwwidgets.NumStepsPerTemp(app_data),
            kwwidgets.RandomComposition(app_data),
        ]
        super().__init__(app_data, output)

        atoms_widget = kwwidgets.GSAtomsWidget(app_data)
        self.register_arg_widget(atoms_widget)
        # Register getting ECI, but we don't actually need a
        # widget for this, it's contained in the app data
        self.clease_args.append(self.get_eci)

        self.register_kwarg_widget(*key_value_widgets)

    def generator_name(self) -> str:
        return "generate_gs_structure"

    def get_eci(self):
        return self.app_data[AppDataKeys.ECI]

    def get_cwd(self):
        return self.app_data[AppDataKeys.CWD]


class ProbeStructure(BaseStructGenWidgets):
    def __init__(self, app_data, output):
        widgets = [
            kwwidgets.AtomsWidget(app_data),
            kwwidgets.ProbeStructTemp(app_data),
            kwwidgets.NumTemp(app_data),
            kwwidgets.NumStepsPerTemp(app_data),
            kwwidgets.ApproxMeanVar(app_data),
        ]
        super().__init__(app_data, output)
        self.register_kwarg_widget(*widgets)

    def generator_name(self) -> str:
        return "generate_probe_structure"



import logging
import random
import time
from collections import Counter
from typing import Tuple, Dict, List, Callable, Sequence
from IPython.display import display, clear_output
import ipywidgets as widgets
import ase
import ase.io
import clease
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar
from clease_gui.colors import bool2symbol
from clease_gui import register_logger, utils

__all__ = ["SupercellDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)


class SupercellDashboard(BaseDashboard):
    def initialize(self):

        self.sc_widgets = []
        for coord in ("x", "y", "z"):
            wdgt = widgets.BoundedIntText(
                description=f"Repetition ({coord}):",
                value=5,
                min=1,
                max=999,
                **self.DEFAULT_STYLE_KWARGS,
            )
            self.sc_widgets.append(wdgt)

        self.make_atoms_btn = widgets.Button(
            description="Make random supercell", layout=self.DEFAULT_BUTTON_LAYOUT
        )
        self.make_atoms_btn.on_click(self._on_make_supercell_click)

        self.desc_label = widgets.Label(value=self.make_desc_label())

        self.view_atoms_btn = utils.make_clickable_button(
            self._on_view_atoms_click,
            description="View Current Supercell",
            layout=self.DEFAULT_BUTTON_LAYOUT,
        )

        self.load_supercell_button = widgets.Button(description="Load Supercell")
        self.load_supercell_text = widgets.Text(description="Filename:", value="supercell.traj")
        self.load_supercell_button.on_click(self._on_load_supercell_click)

        self.save_supercell_button = widgets.Button(description="Save Supercell")
        self.save_supercell_text = widgets.Text(description="Filename:", value="supercell.traj")
        self.save_supercell_button.on_click(self._on_save_supercell_click)

        self.species_output = widgets.Output()
        self.species_output.clear_output()

    def display(self):
        display(self.desc_label)
        vbox_sc_widgets = widgets.VBox(children=self.sc_widgets)

        hbox_buttons = widgets.HBox(
            children=[
                self.make_atoms_btn,
                self.view_atoms_btn,
            ]
        )

        load_hbox = widgets.HBox(children=[self.load_supercell_button, self.load_supercell_text])
        save_hbox = widgets.HBox(children=[self.save_supercell_button, self.save_supercell_text])
        display(save_hbox, load_hbox)
        display(hbox_buttons)
        display(vbox_sc_widgets)
        display(self.species_output)

    def sc_formula_unit(self):
        try:
            supercell = self.supercell
        except ValueError:
            return ""
        else:
            return supercell.get_chemical_formula()

    @update_statusbar
    @utils.disable_cls_widget("load_supercell_button")
    def _on_load_supercell_click(self, b):
        with self.event_context(logger=logger):
            self.load_supercell()

    def load_supercell(self):
        fname = self.load_supercell_text.value
        if fname == "":
            raise ValueError("No filename supplied")
        cwd = self.app_data[self.KEYS.CWD]
        fname = cwd / fname
        self._load_supercell(fname)

    def _load_supercell(self, fname):
        logger.info("Loading supercell from file: %s", fname)
        atoms = ase.io.read(fname)
        self.set_supercell(atoms)
        logger.info("Load successful.")

    def make_random_supercell(self):
        atoms = self.settings.prim_cell.copy()
        atoms *= self.repetitions

        self.settings.set_active_template(atoms)
        nib = [len(x) for x in self.settings.index_by_basis]
        x = self.concentration.get_random_concentration(nib)
        supercell = self._make_atoms_at_conc(x)
        self.set_supercell(supercell)

    @update_statusbar
    @utils.disable_cls_widget("view_atoms_btn")
    def _on_view_atoms_click(self, b):
        with self.event_context(logger=logger):
            self._visualize_supercell()
            self._update_desc_label()

    def _visualize_supercell(self):
        from ase.visualize import view

        supercell = self.supercell
        if supercell is None:
            raise ValueError("No supercell has been made yet.")
        # use a copy to remove any energies, we don't want to see graphs
        # this is a bit of a hack...
        cpy = supercell.copy()
        cpy.calc = None
        view(cpy)
        logger.info("Opened supercell in a new ASE GUI window.")

    def make_desc_label(self):
        text = "Create a supercell."
        try:
            sc = self.supercell
        except ValueError:
            # No supercell present, just return initial text part
            return text
        formula_unit = sc.get_chemical_formula()
        ntot = len(sc)

        text += f" Current formula unit: {formula_unit}."
        text += f" Total number of atoms: {ntot}"

        return text

    def _update_desc_label(self):
        self.desc_label.value = self.make_desc_label()

    def set_supercell(self, atoms: ase.Atoms, draw_species_output=True):
        # XXX: Do we want to set the active template here?
        self.settings.set_active_template(atoms)
        self.app_data[self.KEYS.SUPERCELL] = atoms
        self._update_desc_label()
        if draw_species_output:
            self._make_species_output()

    @property
    def supercell(self):
        try:
            return self.app_data[self.KEYS.SUPERCELL]
        except KeyError:
            raise ValueError("No supercell available in app data.")

    @property
    def repetitions(self) -> Tuple[int]:
        return tuple(wdgt.value for wdgt in self.sc_widgets)

    @update_statusbar
    @utils.disable_cls_widget("make_atoms_btn")
    def _on_make_supercell_click(self, b):
        with self.event_context(logger=logger):
            start_time = time.perf_counter()
            self.make_random_supercell()
            dt = time.perf_counter() - start_time
            logger.info("Supercell created in %.3f s.", dt)

    def _get_new_struct_obj(self):
        """Helper function for creating a NewStructures object"""
        return clease.structgen.NewStructures(self.settings)

    @property
    def concentration(self):
        return self.settings.concentration

    def _make_atoms_at_conc(self, x):
        new_struct = self._get_new_struct_obj()
        num_atoms_in_basis = [len(indices) for indices in self.settings.index_by_basis]
        num_to_insert = self.concentration.conc_in_int(num_atoms_in_basis, x)
        return new_struct._random_struct_at_conc(num_to_insert)

    def save_supercell(self):
        supercell = self.supercell
        fname = self.save_supercell_text.value
        if fname == "":
            raise ValueError("No filename supplied.")
        fname = self.app_data[self.KEYS.CWD] / fname
        logger.info("Saving supercell to file: %s", fname)
        ase.io.write(fname, supercell)
        logger.info("Save successful.")

    @update_statusbar
    @utils.disable_cls_widget("save_supercell_button")
    def _on_save_supercell_click(self, b):
        with self.event_context(logger=logger):
            self.save_supercell()

    def get_species_per_basis(self) -> List[Dict[str, int]]:
        settings = self.settings
        supercell = self.supercell

        idx_by_basis = settings.index_by_basis

        species_per_basis = []
        for idx in idx_by_basis:
            atoms = supercell[idx]
            species = Counter(atoms.symbols)
            species_per_basis.append(species)
        return species_per_basis

    def set_supercell_species(self, species_per_basis: Sequence[Dict[str, int]]):
        # Check that all species per basis add up to the correct number
        # by comparing to the current total
        basis_elements = self.settings.basis_elements
        index_by_basis = self.settings.index_by_basis
        for ii, new_basis in enumerate(species_per_basis):
            cur_count = len(index_by_basis[ii])
            new_count = sum(new_basis.values())
            if cur_count != new_count:
                sublatt = ii + 1
                raise ValueError(
                    (
                        "Incorrect number of elements in sublattice {}." " Expected {}, but got {}."
                    ).format(sublatt, cur_count, new_count)
                )
        # Do the update, now we confirmed the counts are OK!
        atoms = self.supercell.copy()
        for ii, indices in enumerate(index_by_basis):
            new_syms = []
            for elem in basis_elements[ii]:
                new_count = species_per_basis[ii].get(elem, 0)
                new_syms.extend([elem] * new_count)
            assert len(new_syms) == len(indices), (len(new_syms), len(indices))
            # Shuffle sublattice
            random.shuffle(new_syms)
            atoms.symbols[indices] = new_syms

        # Check if we violate concentration constraints. We do not update if that's the case.
        if not self.settings.concentration.is_valid(index_by_basis, atoms):
            raise ValueError("The specified configuration violates the concentration constraints.")
        # We don't need to redraw the species output widgets again
        self.set_supercell(atoms, draw_species_output=False)
        logger.info("Supercell updated.")

    def _make_species_output(self):
        out = self.species_output

        species_per_basis = self.get_species_per_basis()

        update_supercell_btn = widgets.Button(description="Update supercell")

        with out:
            clear_output(wait=True)
            get_species_counts = []
            for ii, species_count in enumerate(species_per_basis):
                sublattice_index = ii + 1
                # Draw the sublattice boxes, and add the callback so we can get number of species
                output, callback = make_sublattice_box(species_count, sublattice_index)
                get_species_counts.append(callback)
                display(output)

            display(update_supercell_btn)

        def make_species_per_basis():
            count = []
            for callback in get_species_counts:
                count_in_sublattice = callback()
                count.append(count_in_sublattice)
            return count

        def on_update_click(b):
            with utils.disable_widget_context(update_supercell_btn):
                with self.event_context(logger=logger):
                    self.set_supercell_species(make_species_per_basis())

        update_supercell_btn.on_click(on_update_click)


def make_sublattice_box(
    species_count: Dict[str, int], sublattice_index: int
) -> Tuple[widgets.Output, Callable[[], Dict[str, int]]]:
    """Draw a box for the sublattice containing a given set of species.
    Returns the Output object with the boxes, as well as a function
    with the counts of the species specified by the user in the boxes in
    that sublattice."""
    # total number of species in the sublattice
    tot_in_basis = sum(species_count.values())

    # The output object in which the boxes are going to be drawn
    sublattice_out = widgets.Output(layout={"border": "1px solid lightgray"})

    def make_make_label_text(is_ok=True):
        s = f"Sublattice {sublattice_index}  {bool2symbol[is_ok]}"
        return f"<h1 style='font-size:17px'> {s} </h1>"

    label = widgets.HTML(value=make_make_label_text())
    # Let's draw the label text on top
    with sublattice_out:
        clear_output()
        display(label)

    # List with all the elemental count boxes
    all_elem_boxes = []
    # A list of species according to the order in all_elem_boxes
    species_lst: List[str] = []

    # A box with the total count
    tot_box = widgets.IntText(
        description="Total in basis:",
        value=tot_in_basis,
        disabled=True,
        style={"description_width": "initial"},
    )
    # Box with the total specified number in this basis
    tot_specified = widgets.IntText(
        description="Total specified:",
        value=0,
        disabled=True,
        style={"description_width": "initial"},
    )
    # Group the "total" boxes horizontally
    tot_hbox = widgets.HBox(children=[tot_box, tot_specified])

    def update():
        """Callback which updates the tot_specified box and the label"""
        new_tot = sum(widget.value for widget in all_elem_boxes)
        tot_specified.value = new_tot
        label.value = make_make_label_text(new_tot == tot_in_basis)

    def observe_change(change):
        """Has the total number of assigned species has changed?"""
        if utils.is_value_change(change):
            update()

    # Create the individual species boxes
    with sublattice_out:
        for species, count in species_count.items():
            wdgt, ratio_wdgt = _make_species_boxes(species, count, tot_in_basis)
            # We need to update tot_specified if we make a change here
            wdgt.observe(observe_change)
            species_lst.append(species)
            all_elem_boxes.append(wdgt)
            hbox = widgets.HBox(children=[wdgt, ratio_wdgt])
            # Draw the count boxes 1-by-1
            display(hbox)

        # Finally, draw the "total" boxes
        display(tot_hbox)
    # Update the count after we made the individual species boxes
    update()

    def get_counts() -> Dict[str, int]:
        """Get the counts of the widgets created by this function.

        Raises a value error if the counts are inconsistent."""
        counts = {}
        for species, wdgt in zip(species_lst, all_elem_boxes):
            value = wdgt.value
            counts[species] = value
        cur_count = sum(counts.values())
        if cur_count != tot_in_basis:
            raise ValueError(
                (
                    "Incorrect number of elements in sublattice {}." " Expected {}, but got {}."
                ).format(sublattice_index, tot_in_basis, cur_count)
            )
        return counts

    return sublattice_out, get_counts


def _make_species_boxes(
    species: str, count: int, tot_in_basis: int
) -> Tuple[widgets.BoundedIntText, widgets.FloatText]:
    """Helper function to make the species count widget and the corresponding ratio box."""
    # The box in which the user types in the species count
    wdgt = widgets.BoundedIntText(value=count, min=0, max=tot_in_basis, description=f"{species}:")
    # A box with the ratio of species this corresponds to
    ratio_wdgt = widgets.FloatText(
        value=count / tot_in_basis,
        description="Ratio:",
        disabled=True,
    )

    def count_changed(change):
        if utils.is_value_change(change):
            ratio_wdgt.value = change.new / tot_in_basis

    wdgt.observe(count_changed)

    return wdgt, ratio_wdgt



from abc import ABC, abstractmethod
from pathlib import Path
import ipywidgets as widgets
from IPython.display import display, clear_output
import numpy as np
from clease import regression
from clease_gui.widget_collection import WidgetCollection, SingleWidget
from clease_gui.style_mixin import WidgetStyleMixin
from clease_gui import utils


class RegressionScheme(ABC, WidgetStyleMixin):
    HYPERPARAMS = None
    NAME = None

    def __init__(self, output=None):
        assert self.HYPERPARAMS is not None, "Set the required hyperparameters!"
        assert self.NAME is not None, "Set the name!"
        self.output = output or widgets.Output()
        self.hyper_widget_dct = {}
        self.hyper_output = widgets.Output()
        self._initialize()

    def display(self):

        # display the hyperparameter dropdown next to the hyperparameter output
        # widget.
        box = widgets.HBox(
            children=[self.hyper_dropdown, self.hyper_output],
            layout={"align_items": "center"},
        )
        with self.output:
            clear_output(wait=True)
            display(box)
            self.hyper_dropdown.value.display()

    def _initialize(self):

        options = []
        for param in self.HYPERPARAMS:
            if param == "alpha":
                widget = Alpha(output=self.hyper_output)
                self.hyper_widget_dct[param] = widget
                options.append((param, widget))
            else:
                raise ValueError(f"Cannot handle parameter: {param}")

        # Make dropdown widget
        dropdown = widgets.Dropdown(
            options=options, description="Hyperparameter:", **self.DEFAULT_STYLE_KWARGS
        )

        def _on_change(change):
            """If we change the dropdown value, we should updatea the
            hyperparameter display"""
            if not utils.is_value_change(change):
                return
            # Display the new selected widget
            change.new.display()

        dropdown.observe(_on_change)
        self.hyper_dropdown = dropdown

    def make_parameters(self):
        params = {}
        for key, widget in self.hyper_widget_dct.items():
            value = widget.get_param()
            params[key] = value

        return params

    @abstractmethod
    def get_fitting_scheme(self):
        """Return the name of the fitting scheme"""

    @abstractmethod
    def get_fitting_scheme_class(self):
        """Return the fitting scheme class"""

    def make_fitting_instance(self, **params):
        """Make an instance of the fitting algorithm
        with a set of parameters"""
        # params = self.make_parameters()
        fit_cls = self.get_fitting_scheme_class()
        return fit_cls(**params)


class LassoScheme(RegressionScheme):
    HYPERPARAMS = ("alpha",)
    NAME = "Lasso"

    def get_fitting_scheme(self):
        return "lasso"

    def get_fitting_scheme_class(self):
        return regression.Lasso


class L2Scheme(RegressionScheme):
    HYPERPARAMS = ("alpha",)
    NAME = "L2"

    def get_fitting_scheme(self):
        return "l2"

    def get_fitting_scheme_class(self):
        return regression.Tikhonov


class FloatHyperparameter(WidgetStyleMixin):
    # Name of the hyperparameter to be passed
    # into a regression scheme
    FIXED = "fixed_hyper"
    OPTIMIZE = "optimize_hyper"

    # Settings in the optimization boxes
    MIN_SETTINGS = {
        "value": 0.01,
        "min": 0,
        "max": 9999999,
    }
    MAX_SETTINGS = {
        "value": 1,
        "min": 0,
        "max": 9999999,
    }
    STEP_SETTINGS = {
        "value": 25,
        "min": 0,
        "max": 9999999,
    }
    DEFAULT_VALUE = 0.01  # Default float hyperparam value

    def __init__(self, output=None):
        super().__init__()
        self.output = output or widgets.Output()
        self._initialize()

    def display(self):
        box = widgets.VBox(
            children=[
                self.optimize_box,
                self.float_box,
                self.spacing_dropdown,
                self.range_boxes,
            ]
        )
        with self.output:
            clear_output(wait=True)
            display(box)

    @property
    def mode(self):
        value = self.optimize_box.value
        if value:
            return self.OPTIMIZE
        return self.FIXED

    def _initialize(self):
        """Construct all relevant widgets"""
        self.optimize_box = widgets.Checkbox(
            value=False,
            description="Optimize Hyperparameter?",
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.float_box = widgets.FloatText(
            value=self.DEFAULT_VALUE,
            description="Value:",
            **self.DEFAULT_STYLE_KWARGS,
        )
        # Boxes for optimization, disabled by default
        self.spacing_dropdown = widgets.Dropdown(
            options=[
                ("Linear Spacing", "linspace"),
                ("Logarithmic Spacing", "logspace"),
            ],
            value="logspace",
            description="Spacing type:",
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.range_boxes = make_spacing_boxes(
            self.MIN_SETTINGS,
            self.MAX_SETTINGS,
            self.STEP_SETTINGS,
            style_kwargs=self.DEFAULT_STYLE_KWARGS,
        )

        def on_opt_change(change):
            if not utils.is_value_change(change):
                return
            self.update_mode_widgets()

        self.update_mode_widgets()  # Update the first time

        self.optimize_box.observe(on_opt_change)

    @property
    def is_optimization(self):
        return self.mode == self.OPTIMIZE

    def update_mode_widgets(self) -> None:
        is_opt = self.is_optimization
        is_fixed = not is_opt

        self.float_box.disabled = is_opt
        self.spacing_dropdown.disabled = is_fixed
        for box in self.range_boxes.children:
            box.disabled = is_fixed

    def get_min_max_step(self):
        min_val = self.range_boxes.children[0].value
        max_val = self.range_boxes.children[1].value
        step_val = self.range_boxes.children[2].value
        return (min_val, max_val, step_val)

    def get_spacing_type(self):
        return self.spacing_dropdown.value

    def get_float_value(self):
        return self.float_box.value

    def get_param(self) -> np.ndarray:
        mode = self.mode
        if mode == self.FIXED:
            # return value in a 1 elem np array
            return np.array([self.get_float_value()])

        # Optimize, generate the spacing
        min_val, max_val, step_val = self.get_min_max_step()
        spacing_type = self.get_spacing_type()
        # Get the corresponding function from NumPy
        func = getattr(np, spacing_type)
        if spacing_type == "logspace":
            # Logspace is in base 10 by default
            # Convert min and max to the corresponding log values
            min_val = np.log10(min_val)
            max_val = np.log10(max_val)
        return func(min_val, max_val, step_val)


class Alpha(FloatHyperparameter):
    DEFAULT_VALUE = 0.01
    MIN_SETTINGS = {
        "value": 1e-5,
        "min": 0,
        "max": 9999999,
    }
    MAX_SETTINGS = {
        "value": 1.0,
        "min": 0,
        "max": 9999999,
    }
    STEP_SETTINGS = {
        "value": 25,
        "min": 0,
        "max": 9999999,
    }


class ScoringSchemeWidget(WidgetCollection):
    """Widget to select the scoring scheme,
    and optionally the number of splits"""

    def make_widgets(self):
        self.scoring_scheme = widgets.Dropdown(
            options=[
                ("K-fold", "k-fold"),
                ("Leave-one-out CV", "loocv"),
            ],
            value="loocv",
            description="Scoring scheme:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.nsplits = widgets.IntText(
            value=10,
            description="Number of splits:",
            **self.DEFAULT_STYLE_KWARGS,
            disabled=self.scoring_scheme.value == "loocv",
        )

        def on_scheme_change(change):
            if utils.is_value_change(change):
                # Disable n-splits if scoring scheme is loocv
                self.nsplits.disabled = change.new == "loocv"

        self.scoring_scheme.observe(on_scheme_change)
        widget_box = widgets.HBox(children=[self.scoring_scheme, self.nsplits])

        return [SingleWidget("WidgetBox", widget_box)]

    @property
    def value(self):
        return dict(scoring_scheme=self.scoring_scheme.value, nsplits=self.nsplits.value)


class DBSelectWidget(WidgetCollection):
    """Optional select a database to fit ECI's to."""

    def make_widgets(self):
        # Store the text widget in the object, to make it easier to access
        self.db_select = widgets.Text(
            description="Select a database:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        label = widgets.Label("Leave empty to use the database specified in the main settings.")

        hbox = widgets.HBox(children=[self.db_select, label])
        # return dict(db_select=hbox)
        return [SingleWidget("db_select", hbox)]

    @property
    def value(self):
        value = self.db_select.value
        if value == "":
            return None
        return Path(value)


def make_spacing_boxes(min_dct: dict, max_dct: dict, steps_dct: dict, style_kwargs=None):
    """Heplper function to make steps boxes"""
    style_kwargs = style_kwargs or {}

    range_min = widgets.BoundedFloatText(
        description="min:",
        **min_dct,
        **style_kwargs,
    )
    range_max = widgets.BoundedFloatText(
        description="max:",
        **max_dct,
        **style_kwargs,
    )
    range_steps = widgets.BoundedIntText(
        description="steps:",
        **steps_dct,
        **style_kwargs,
    )

    range_box = widgets.VBox(
        children=[range_min, range_max, range_steps],
        layout=dict(justify_content="space-between"),
    )
    return range_box



from .concentration import *
from .structure_settings import *
from .cluster_dashboard import *
from .ce_settings import *
from .settings_maker import *

__all__ = (
    concentration.__all__
    + structure_settings.__all__
    + settings_maker.__all__
    + cluster_dashboard.__all__
    + ce_settings.__all__
)



import logging
import functools
from pathlib import Path, PurePath
from contextlib import contextmanager
from ipywidgets import DOMWidget
import ipywidgets as widgets
from clease_gui.logging_widget import register_logger

__all__ = [
    "is_value_change",
    "disable_widget_context",
    "disables_widget",
    "disable_cls_widget",
    "make_clickable_button",
    "get_assets_path",
]

logger = logging.getLogger(__name__)
register_logger(logger)


def get_assets_path() -> PurePath:
    """Return the path to the assets directory"""
    return Path(__file__).parent / "assets"


def make_clickable_button(*click_event, description="", **button_kwargs) -> widgets.Button:
    """Helper function for making a new button and adding 1 or more callback
    functions on click events.
    Combines the general pattern of

    button = widgets.Button(**kwargs)
    button.on_click(my_event)

    into a single function call.
    """
    button = widgets.Button(description=description, **button_kwargs)
    for callback in click_event:
        button.on_click(callback)
    return button


def is_value_change(change: dict) -> bool:
    """Determine if a widget change is due to a change in value.
    This change comes from the widget.observe() function.
    """
    return change["type"] == "change" and change["name"] == "value"


@contextmanager
def disable_widget_context(widget: DOMWidget):
    """Context manager which temporarily disables a widget.
    Usage:

    with disable_widget_context(my_widget):
        do_stuff()
    """
    if not isinstance(widget, DOMWidget):
        raise TypeError(f"Widget must be of type DOMWidget, got {type(widget)}")
    # Get the current state of the widget
    prev_state: bool = widget.disabled
    widget.disabled = True
    logger.debug("Widget disabled: %s", widget.disabled)
    try:
        yield
    finally:
        # Enable the widget again, if it was previously enabled.
        widget.disabled = prev_state
        logger.debug("Widget restored, new state: %s", widget.disabled)


def disables_widget(widget):
    """Decorator which disables a widget while function is executing.
    Usage:
    @disables_widget(my_widget)
    def my_func():
        # While this function is executing, the widget "my_widget" is disabled.
        do_stuff()
    """

    def decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            with disable_widget_context(widget):
                return func(*args, **kwargs)

        return _wrapper

    return decorator


def disable_cls_widget(widget_name: str):
    """Decorator which disables a widget of a class which is a member of "self"
    Assumes the widget name exists in "self", i.e. it uses getattr(self, widget_name)
    to access the widget.

    @disable_cls_widget('my_widget_name')
    def my_func(self):
        do_stuff()
    """

    def decorator(func):
        @functools.wraps(func)
        def _wrapper(self, *args, **kwargs):
            # Find the widget to be disabled.
            widget = getattr(self, widget_name)
            # Disable the widget during function execution.
            with disable_widget_context(widget):
                return func(self, *args, **kwargs)

        return _wrapper

    return decorator



from typing import Sequence
import logging
from abc import ABC
from IPython.display import display, clear_output
import ipywidgets as widgets
import numpy as np
from clease_gui import utils, register_logger
from clease_gui.app_data import Notification, AppDataKeys
from clease_gui.widget_collection import WidgetCollection, SingleWidget

logger = logging.getLogger(__name__)
register_logger(logger)


class MaxClusterDiaWidget(SingleWidget):
    """Special widget for max cluster diameters, it needs to update itself,
    and recreate the VBox every time it needs to update."""

    DEFAULT_MCD_VALUE = 5.0

    def __init__(self, name, num_boxes, priority=0):
        self.mcd_boxes = []
        self.label = widgets.Label(
            value="Max cluster diameter:",
            # # layout=self.widget.layout,
            # display='flex',
            # flex_flow='column',
        )

        # Update the number of boxes, and create the initial widget
        self._update_boxes(num_boxes)
        initial_widget = self.make_widget()
        super().__init__(name, initial_widget, priority=priority)
        # The output object we will be drawing the MCD object on
        self.output = widgets.Output()

    def make_widget(self):
        return widgets.VBox(children=self.mcd_boxes)

    def update_widget(self, num_boxes):
        self._update_boxes(num_boxes)
        self.widget = self.make_widget()
        # The output has already been displayed, so it doesn't need to be updated
        self.display_widget(display_output=False)

    @property
    def value(self):
        # Return a NumPy array with the MCD values
        return np.array([child.value for child in self.widget.children])

    def _update_boxes(self, num_boxes):
        def add_box():
            """Helper function for making a bound float box"""
            i = len(self.mcd_boxes)
            box = widgets.BoundedFloatText(
                description=f"{i+2}-body",
                value=self.DEFAULT_MCD_VALUE,
                min=0,
                disabled=False,
            )
            self.mcd_boxes.append(box)

        delta = len(self.mcd_boxes) - num_boxes

        if delta > 0:
            # We have more boxes than MCS, pop the difference
            for _ in range(delta):
                try:
                    self.mcd_boxes.pop()
                except IndexError:
                    # mcd_boxes is empty, no need to continue
                    break
        elif delta < 0:
            # We need to add boxes
            for _ in range(abs(delta)):
                add_box()
        logger.debug("Number of boxes in MCD object: %s", len(self.mcd_boxes))

    def set_value(self, value):
        logger.debug("Updating max_cluster_dia values with %s", value)
        if not len(value) == len(self.mcd_boxes):
            raise ValueError(
                "Incorrect length of value. Got {}, expected {}",
                len(value),
                len(self.mcd_boxes),
            )
        for ii, val in enumerate(value):
            self.mcd_boxes[ii].value = val

    def display_widget(self, display_output=True):
        """Display the widget.

        :param display_output: call display() on the output oject.
            Should only be True on the first call, to avoid repeated objects, unless it needs
            to be redrawn after a clear_output()
        """
        with self.output:
            # We are the only ones drawing in this output,
            # so we always clear on an update
            clear_output(wait=True)
            vbox = widgets.VBox(children=[self.label, self.widget])
            vbox.layout.align_items = "center"
            display(vbox)
            # display(self.label)
            # display(self.widget)
        if display_output:
            logger.debug("Drawing mcd widget output")
            # We don't always want to redraw the entire output, e.g. when simply redrawing
            # this widget in its own output
            display(self.output)


class MaxClusterWidget(WidgetCollection):
    """Widget for doing max cluster size and max cluster diameter"""

    MAX_CLUSTER_SIZE_NAME = "max_cluster_size"
    MAX_CLUSTER_DIA_NAME = "max_cluster_dia"

    def make_widgets(self):
        widget_lst = []

        self._max_cluster_output = widgets.Output()
        # Max cluster size and max cluser diameter
        self.max_cluster_size = widgets.BoundedIntText(
            value=4,
            min=2,
            description="Max cluster size:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        # This is now the only widget which returns any information
        # on cluster sizes, since MCD now uniquely determines the
        # number of clusters since version 0.10.6.
        # max_cluster_size is inferred from the length of MCD.
        self.mcd_widget = MaxClusterDiaWidget(
            self.MAX_CLUSTER_DIA_NAME, self.number_of_boxes(), priority=9
        )

        widget_lst.append(self.mcd_widget)

        def on_max_cluster_size_change(change):
            if utils.is_value_change(change):
                self.update_mcd_widget()

        self.max_cluster_size.observe(on_max_cluster_size_change)
        return widget_lst

    def update_mcd_widget(self):
        logger.debug("Updating mcd widget")
        self.mcd_widget.update_widget(self.number_of_boxes())

    def number_of_boxes(self):
        """Helper function which calculates how many MCD
        # boxes we should have"""
        return self.max_cluster_size.value - 1

    def display_single_widget(self, widget: SingleWidget):
        # Display the max cluster dia and max cluster size widget together
        if widget.name == self.MAX_CLUSTER_SIZE_NAME:
            # We display both when calling max_cluster_dia
            return
        if widget.name == self.MAX_CLUSTER_DIA_NAME:
            box = widgets.HBox(children=[self.max_cluster_size, self.mcd_widget.output])
            box.layout.align_items = "baseline"
            self.mcd_widget.display_widget(display_output=False)
            with self._max_cluster_output:
                clear_output()
                display(box)
            display(self._max_cluster_output)
            return
        super().display_single_widget(widget)

    def get_max_cluster_size(self):
        return self.max_cluster_size.value

    def set_widget_value(self, key, value):
        if key == "max_cluster_size":
            self.max_cluster_size.value = value
            self.update_mcd_widget()
        elif key == "max_cluster_dia":
            self.mcd_widget.set_value(value)
        else:
            super().set_widget_value(key, value)


class CommonSettingsButtons(MaxClusterWidget, WidgetCollection, ABC):
    """Some buttons which are shared between all the CLEASE settings types,
    most notably the max cluster dia and max cluster size"""

    def __init__(self, app_data, *args, **kwargs):
        # Store the FloatBoxes for the max cluster dia array
        self.app_data = app_data
        super().__init__(*args, **kwargs)

    def make_widgets(self):
        widget_lst = super().make_widgets()

        # Basis func type
        basis_func_type = widgets.Dropdown(
            options=[
                ("Polynomial", "polynomial"),
                ("Trigonometric", "trigonometric"),
                ("Binary Linear", "binary_linear"),
            ],
            description="Basis function:",
            **self.DEFAULT_STYLE_KWARGS,
        )
        widget_lst.append(SingleWidget("basis_func_type", basis_func_type, priority=100))

        # include background atoms
        self.include_background_atoms = widgets.Dropdown(
            options=[("True", True), ("False", False)],
            value=False,
            description="Background atoms:",
            **self.DEFAULT_STYLE_KWARGS,
        )
        # Ensure that include_background_atoms is only enabled when no settings are present
        self.app_data.subscribe(AppDataKeys.SETTINGS, self._on_settings_change)
        widget_lst.append(
            SingleWidget("include_background_atoms", self.include_background_atoms, priority=5)
        )

        # db_name
        db_name = widgets.Text(
            value="clease.db",
            description="Database name:",
            **self.DEFAULT_STYLE_KWARGS,
        )
        widget_lst.append(SingleWidget("db_name", db_name, priority=1))
        return widget_lst

    def get_widget_value(self, widget: SingleWidget):
        if widget.name == "db_name":
            # Add the cwd to the database name
            # cwd = self.app_data[gui.AppDataKeys.CWD]
            # return str(cwd / widget.widget.value)
            return widget.widget.value
        return super().get_widget_value(widget)

    def _on_settings_change(self, change: Notification) -> None:
        if change.key == AppDataKeys.SETTINGS:
            # Update is_background_atoms widget state
            is_disabled = change.new_value is not None
            logger.debug("Updating is_background_atoms to: %s", is_disabled)
            self.include_background_atoms.disabled = is_disabled


class CrystalButtons(WidgetCollection, ABC):
    def __init__(self, app_data, *args, **kwargs):
        # Store the FloatBoxes for the max cluster dia array
        self.app_data = app_data
        super().__init__(*args, **kwargs)

    def make_widgets(self) -> Sequence[SingleWidget]:
        return []


class CEBulkButtons(CrystalButtons):
    def make_widgets(self):
        widget_list = super().make_widgets()

        crystalstructure = widgets.Dropdown(
            options=[
                ("Simple Cubic", "sc"),
                ("FCC", "fcc"),
                ("BCC", "bcc"),
                ("HCP", "hcp"),
                ("Diamond", "diamond"),
                ("Zincblende", "zincblende"),
                ("Rocksalt", "rocksalt"),
                ("Cesium Chloride", "cesiumchloride"),
                ("Fluorite", "fluorite"),
                ("Wurtzite", "wurtzite"),
            ],
            description="Crystal Structure:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        a = widgets.BoundedFloatText(
            value=None,
            min=0,
            max=999,
            description="a (Å):",
            description_tooltip=("Setting the value to 0 corresponds to setting it to None"),
            **self.DEFAULT_STYLE_KWARGS,
        )

        c = widgets.BoundedFloatText(
            value=None,
            min=0,
            max=999,
            description="c (Å):",
            description_tooltip=("Setting the value to 0 corresponds to setting it to None"),
            **self.DEFAULT_STYLE_KWARGS,
        )

        covera = widgets.BoundedFloatText(
            value=None,
            min=0,
            max=999,
            description="c/a:",
            description_tooltip=("Setting the value to 0 corresponds to setting it to None"),
            **self.DEFAULT_STYLE_KWARGS,
        )

        widget_list.extend(
            [
                SingleWidget("crystalstructure", crystalstructure, priority=80),
                SingleWidget("a", a, priority=75),
                SingleWidget("c", c, priority=70),
                SingleWidget("covera", covera, priority=65),
            ]
        )
        return widget_list

    def get_widget_value(self, widget: SingleWidget):
        """Helper function for getting the value from a button."""
        if widget.name in ["a", "c", "covera"]:
            # Set value to 0 to ignore
            value = widget.widget.value
            if value == 0:
                return None
        return super().get_widget_value(widget)


class CECrystalButtons(CrystalButtons):
    def make_widgets(self):
        widget_list = super().make_widgets()
        spacegroup = widgets.IntText(
            value=1,
            description="Spacegroup:",
            **self.DEFAULT_STYLE_KWARGS,
        )

        widget_list.append(SingleWidget("spacegroup", spacegroup, priority=80))
        return widget_list


class CESlabButtons(CrystalButtons):
    def make_widgets(self):
        widget_list = super().make_widgets()

        millers = []
        coord = ["x", "y", "z"]
        for c in coord:
            millers.append(
                widgets.IntText(
                    value=4,
                    description=f"Miller ({c}):",
                    **self.DEFAULT_STYLE_KWARGS,
                )
            )
        miller_box = widgets.VBox(children=millers)

        widget_list.append(SingleWidget("miller", miller_box, priority=80))
        return widget_list

    def get_widget_value(self, widget: SingleWidget):
        """Helper function for getting the value from a button."""
        if widget.name == "miller":
            value = np.zeros(3)
            wdgt = widget.widget
            for ii in range(3):
                value[ii] = wdgt.children[ii].value
            return value
        return super().get_widget_value(widget)


class FixedSizeMode(WidgetCollection):
    def make_widgets(self):
        widgets_list = []
        self.cell_size = widgets.Text(
            description="Cell size:",
            value="3, 3, 3",
            **self.DEFAULT_STYLE_KWARGS,
        )

        label_msg = """Example: "1, 2, 3" for (1, 2, 3) or "3" for (3, 3, 3). """
        label = widgets.Label(value=label_msg)

        box = widgets.HBox(children=[self.cell_size, label])

        widgets_list.append(SingleWidget("size", box))
        return widgets_list

    def get_widget_value(self, widget: SingleWidget):
        if widget.name == "size":
            cell_widget = widget.widget.children[0]
            return self._parse_text(cell_widget.value)
        return super().get_widget_value(widget)

    def _parse_text(self, text):
        orig_text = str(text)
        text = text.strip()
        text = text.replace(",", " ")
        parts = text.split()
        if len(parts) == 1:
            rep = int(parts[0])
            return tuple(rep for _ in range(3))
        elif len(parts) == 3:
            return tuple(map(int, parts))
        else:
            raise ValueError("Cannot determine size from string: {}".format(orig_text))

    def set_widget_value(self, key, value):
        if key == "size":
            self.cell_size.value = value
        else:
            super().set_widget_value(key, value)


class SupercellFactorSizeMode(WidgetCollection):
    def make_widgets(self):
        widgets_list = []
        supercell_factor = widgets.BoundedIntText(
            description="Supercell Factor:",
            value=27,
            min=0,
            max=999,
            **self.DEFAULT_STYLE_KWARGS,
        )

        widgets_list.append(SingleWidget("supercell_factor", supercell_factor, priority=50))

        skew_threshold = widgets.BoundedIntText(
            description="Skew Treshold:",
            value=40,
            min=0,
            max=999,
            **self.DEFAULT_STYLE_KWARGS,
        )
        widgets_list.append(SingleWidget("skew_threshold", skew_threshold, priority=50))
        return widgets_list



import os
import click
from .main_cli import clease_gui_cli

__all__ = ["launch"]


def _build_cmd(args, notebook):
    if notebook:
        cmd = "jupyter notebook"
    else:
        cmd = "jupyter-lab"
    if args:
        joined = " ".join(args)
        cmd = f"{cmd} {joined}"
    return cmd


@clease_gui_cli.command()
@click.option(
    "-n",
    "--notebook",
    is_flag=True,
    help="""
Use Jupyter notebook instead of Jupyter lab?.""",
)
@click.argument("args", nargs=-1)
def launch(args, notebook):
    cmd = _build_cmd(args, notebook)
    os.system(cmd)



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease.structgen import NewStructures
from clease_gui import register_logger, utils
from clease_gui.base_dashboard import BaseDashboard
from clease_gui.status_bar import update_statusbar
from . import StructureGenerationDashboard, GeneralNewStructSettingsDashboard

__all__ = ["NewStructureDashboard"]

logger = logging.getLogger(__name__)
register_logger(logger)

# Register the clease structgen logger
logger_clease = logging.getLogger("clease.structgen.new_struct")
register_logger(logger_clease)


class NewStructureDashboard(BaseDashboard):
    """Main dashboard for generating structures"""

    def initialize(self):

        self.general_settings_out = widgets.Output()
        self.structure_gen_out = widgets.Output()

        self.general_new_struct_dashboard = GeneralNewStructSettingsDashboard(self.app_data)
        with self.general_settings_out:
            clear_output()
            self.general_new_struct_dashboard.display()

        self.struct_gen_dashboard = StructureGenerationDashboard(self.app_data)
        with self.structure_gen_out:
            clear_output()
            self.struct_gen_dashboard.display()

        self.tab = widgets.Tab(
            children=[
                self.general_settings_out,
                self.structure_gen_out,
            ]
        )

        self.tab.set_title(0, "General Settings")
        self.tab.set_title(1, "Structure Generation")

        # Button for running the new structures generation
        self.generate_widget = utils.make_clickable_button(
            self._on_generate_click,
            description="Generate",
            button_style="primary",
            tooltip=("Generate new structures, and place them directly " "into the database."),
        )

        # Make a box which places the button centralized
        box_layout = widgets.Layout(
            display="flex", flex_flow="column", align_items="center", width="100%"
        )
        self.generate_box = widgets.HBox(children=[self.generate_widget], layout=box_layout)

    def display(self):
        display(self.tab)
        display(self.generate_box)

    def _make_new_struct_instance(self):
        dash = self.general_new_struct_dashboard  # alias
        generation_number = dash.generation_number
        struct_per_gen = dash.struct_per_gen

        logger.debug(
            "Generation number: %s, struct per gen: %s",
            generation_number,
            struct_per_gen,
        )

        settings = self.app_data.get(self.KEYS.SETTINGS, None)
        if settings is None:
            logger.error("Settings does not exist yet.")
            return None

        new_struct = NewStructures(
            settings,
            generation_number=generation_number,
            struct_per_gen=struct_per_gen,
        )
        return new_struct

    def get_generator_settings(self):
        """Collect the args and kwargs from the generation dashboard"""
        dash = self.struct_gen_dashboard  # Alias

        args = dash.get_args()
        kwargs = dash.get_kwargs()
        func_name = dash.get_generator_name()

        return dict(func_name=func_name, args=args, kwargs=kwargs)

    def _generate(self, new_struct, generator_settings):
        """Run the generate function on the new struct class"""
        func_name = generator_settings["func_name"]
        args = generator_settings["args"]
        kwargs = generator_settings["kwargs"]

        # Execute the generation function
        logger.debug("Executing generation function: %s", func_name)
        logger.debug("Generator args: %s", args)
        logger.debug("Generator kwargs: %s", kwargs)
        getattr(new_struct, func_name)(*args, **kwargs)

    @update_statusbar
    def _on_generate_click(self, b):
        with utils.disable_widget_context(self.generate_widget):
            with self.event_context(logger=logger):
                # Disable the generate more structures button
                self._do_generate()

    def _do_generate(self):
        logger.info("Generating new structures")
        generator_settings = self.get_generator_settings()
        logger.debug("Running with the following settings:\n%s", generator_settings)
        new_struct = self._make_new_struct_instance()
        if new_struct is None:
            raise RuntimeError("Failed to make a NewStructures instance.")
        self._generate(new_struct, generator_settings)

        logger.info("Structure generation complete!")
